#region Copyright information
// 
// Copyright  2005-2013 Yongkee Cho. All rights reserved.
// 
// This code is a part of the Biological Object Library and governed under the terms of the
// GNU Lesser General  Public License (LGPL) version 2.1 which accompanies this distribution.
// For more information on the LGPL, please visit http://bol.codeplex.com/license.
// 
// - Filename: Matrix.cs
// - Author: Yongkee Cho
// - Email: yongkeecho@gmail.com
// - Date Created: 2012-11-06 12:13 PM
// - Last Modified: 2013-01-25 3:59 PM
// 
#endregion
using System;
using System.Collections.Generic;
using System.Linq;

namespace BOL.Collections.Generics
{
    /// <summary>
    /// Represents a Matrix generic class.
    /// </summary>
    public class Matrix<TRowKey, TColumnKey, TElement> : ICloneable, IEquatable<Matrix<TRowKey, TColumnKey, TElement>>
    {
        #region Public properties

        /// <summary>Returns the number of rows.</summary>
        public int M { get { return Elements.Length; } }

        /// <summary>Returns the number of columns.</summary>
        public int N { get { return Elements[0].Length; } }

        public TRowKey[] RowKeys { get; private set; }

        public TColumnKey[] ColumnKeys { get; private set; }

        public TElement[][] Elements { get; private set; }

        /// <summary>Returns the row vectors.</summary>
        public IEnumerable<IGrouping<TRowKey, TElement>> Rows
        {
            get
            {
                for (var i = 0; i < M; i++)
                    yield return new Vector<TRowKey, TElement>(RowKeys[i], Elements[i]);
            }
        }

        /// <summary>Returns the column vectors.</summary>
        public IEnumerable<IGrouping<TColumnKey, TElement>> Columns
        {
            get
            {
                for (var j = 0; j < N; j++)
                {
                    var column = new TElement[M];
                    for (var i = 0; i < M; i++)
                        column[i] = Elements[i][j];

                    yield return new Vector<TColumnKey, TElement>(ColumnKeys[j], column);
                }
            }
        }

        #endregion

        #region Constructors

        public Matrix(TRowKey[] rowKeys, TColumnKey[] columnKeys, TElement[][] elements)
        {
            if (rowKeys == null)
                throw new ArgumentNullException("rowKeys");
            if (columnKeys == null)
                throw new ArgumentNullException("columnKeys");
            if (elements == null)
                throw new ArgumentNullException("elements");

            RowKeys = rowKeys;
            ColumnKeys = columnKeys;
            Elements = elements;
        }

        public Matrix(TRowKey[] rowKeys, TElement[][] elements)
            : this(rowKeys, Enumerable.Repeat(default(TColumnKey), elements[0].Length).ToArray(), elements) { }

        public Matrix(TColumnKey[] columnKeys, TElement[][] elements)
            : this(Enumerable.Repeat(default(TRowKey), elements.Length).ToArray(), columnKeys, elements) { }

        public Matrix(TElement[][] elements)
            : this(Enumerable.Repeat(default(TRowKey), elements.Length).ToArray(), Enumerable.Repeat(default(TColumnKey), elements[0].Length).ToArray(), elements) { }

        #endregion
        
        #region ICloneable implementation

        public Matrix<TRowKey, TColumnKey, TElement> Clone()
        {
            return new Matrix<TRowKey, TColumnKey, TElement>(
                RowKeys.Select(x => x).ToArray(), 
                ColumnKeys.Select(x => x).ToArray(), 
                Elements.Select(x => x.Select(y => y).ToArray()).ToArray());
        }

        object ICloneable.Clone()
        {
            return Clone();
        }

        #endregion

        #region IEquatable<Matrix<TRowKey, TColumnKey, TElement>> interface

        public bool Equals(Matrix<TRowKey, TColumnKey, TElement> other)
        {
            if (M != other.M || N != other.N)
                return false;

            for (var i = 0; i < M; i++)
                for (var j = 0; j < N; j++)
                    if (!Elements[i][j].Equals(other.Elements[i][j]))
                        return false;

            return true;
        }

        #endregion

        #region Object overriden

        public override int GetHashCode()
        {
            return Rows.Aggregate(1, (hashCode, current) => hashCode ^ current.GetHashCode());
        }

        public override bool Equals(object other)
        {
            if (other == null)
                throw new NullReferenceException();

            if (!(other is Matrix<TRowKey, TColumnKey, TElement>))
                throw new InvalidCastException("The 'other' argument is not a Matrix<TRowKey, TColumnKey, TElement> object.");

            return Equals(other as Matrix<TRowKey, TColumnKey, TElement>);
        }

        public override string ToString()
        {
            return String.Format("Matrix[{0}x{1}]", M, N);
        }

        #endregion
    }
}
