#region Copyright information
// 
// Copyright  2005-2013 Yongkee Cho. All rights reserved.
// 
// This code is a part of the Biological Object Library and governed under the terms of the
// GNU Lesser General  Public License (LGPL) version 2.1 which accompanies this distribution.
// For more information on the LGPL, please visit http://bol.codeplex.com/license.
// 
// - Filename: MatrixReader.cs
// - Author: Yongkee Cho
// - Email: yongkeecho@gmail.com
// - Date Created: 2012-09-06 11:39 AM
// - Last Modified: 2013-01-25 3:59 PM
// 
#endregion
using System;
using System.IO;
using BOL.Collections.Generics;

namespace BOL.IO
{
    public class MatrixReader<TElement> : ReaderBase<Matrix<string, string, TElement>>
    {
        #region Private variables

        private readonly char[] _dataDelimiter;
        private readonly string _commentDelimiter;
        private readonly int _numberOfRowLabels;
        private readonly int _numberOfColumnLabels;
        private readonly Converter<string, TElement> _converter;

		#endregion

        #region Constructors

        public MatrixReader(Stream stream, char[] dataDelimiter, string commentDelimiter, int numberOfLabelColumns, int numberOfLabelRows, Converter<string, TElement> converter)
    		:base(stream)
        {
            _dataDelimiter = dataDelimiter;
            _commentDelimiter = commentDelimiter;
            _numberOfRowLabels = numberOfLabelColumns;
            _numberOfColumnLabels = numberOfLabelRows;
            _converter = converter;
		}

        public MatrixReader(Stream stream, char[] dataDelimiter, int numberOfLabelColumns, int numberOfLabelRows, Converter<string, TElement> converter)
            : this(stream, dataDelimiter, null, numberOfLabelColumns, numberOfLabelRows, converter) { }

        public MatrixReader(string path, char[] dataDelimiter, string commentDelimiter, int numberOfLabelColumns, int numberOfLabelRows, Converter<string, TElement> converter)
            : this(new FileStream(path, FileMode.Open), dataDelimiter, commentDelimiter, numberOfLabelColumns, numberOfLabelRows, converter) { }

        public MatrixReader(string path, char[] dataDelimiter, int numberOfLabelColumns, int numberOfLabelRows, Converter<string, TElement> converter)
            : this(new FileStream(path, FileMode.Open), dataDelimiter, null, numberOfLabelColumns, numberOfLabelRows, converter) { }

        public MatrixReader(Stream stream, char[] dataDelimiter, string commentDelimiter, Converter<string, TElement> converter) : this(stream, dataDelimiter, commentDelimiter, 0, 0, converter) { }

        public MatrixReader(Stream stream, char[] dataDelimiter, Converter<string, TElement> converter) : this(stream, dataDelimiter, null, 0, 0, converter) { }

        public MatrixReader(string path, char[] dataDelimiter, string commentDelimiter, Converter<string, TElement> converter) : this(path, dataDelimiter, commentDelimiter, 0, 0, converter) { }

        public MatrixReader(string path, char[] dataDelimiter, Converter<string, TElement> converter) : this(path, dataDelimiter, null, 0, 0, converter) { }

		#endregion

        #region IReader<T> implementation

        public override Matrix<string, string, TElement> Read()
		{
            string line;

            string[] temp; // temporary string array splitted by delimiter
            TElement[][] data;
            int i, j; // row, column indices
            var rowLabels = new string[_numberOfRowLabels][];
            var columnLabels = new string[_numberOfColumnLabels][];
            int m = 0, n = 0, dataStartRow = 0; // number of data rows/columns

            if (!String.IsNullOrEmpty(_commentDelimiter))
            {
                var commentLength = _commentDelimiter.Length;
                line = StreamReader.ReadLine();
                dataStartRow = 0;

                while (line.StartsWith(_commentDelimiter) || line == String.Empty)
                    dataStartRow++; 

                for (i = 0; i < _numberOfColumnLabels; i++)
                {
                    temp = line.Split(_dataDelimiter);
                    n = temp.Length - _numberOfRowLabels;
                    columnLabels[i] = new string[n];
                    for (j = 0; j < n; j++)
                        columnLabels[i][j] = temp[j + _numberOfRowLabels].Trim();
                    dataStartRow++;
                }
            }
            else
            {
                for (i = 0; i < _numberOfColumnLabels; i++)
                {
                    line = StreamReader.ReadLine();
                    temp = line.Split(_dataDelimiter);
                    n = temp.Length - _numberOfRowLabels;
                    columnLabels[i] = new string[n];
                    for (j = 0; j < n; j++)
                        columnLabels[i][j] = temp[j + _numberOfRowLabels].Trim();

                    dataStartRow++;
                }
            }

            // count the number of rows
            while (!StreamReader.EndOfStream)
            {
                StreamReader.ReadLine();
                m++;
            }
            data = new TElement[m][];
            for (i = 0; i < _numberOfRowLabels; i++)
                rowLabels[i] = new string[m];

            StreamReader.BaseStream.Seek(0, SeekOrigin.Begin);
            for (i = 0; i < dataStartRow; i++)
                StreamReader.ReadLine(); // pass header line

            line = StreamReader.ReadLine();

            // estimate n when no column label
            if (dataStartRow == 0)
            {
                temp = line.Split(_dataDelimiter);
                n = temp.Length - _numberOfRowLabels;
            }

            for (i = 0; i < m; i++)
            {
                temp = line.Split(_dataDelimiter);

                // assign row labels
                for (j = 0; j < _numberOfRowLabels; j++)
                    rowLabels[j][i] = temp[j].Trim();

                // assign data
                data[i] = new TElement[n];
                for (j = 0; j < n; j++)
                    data[i][j] = _converter(temp[j + _numberOfRowLabels]);

                line = StreamReader.ReadLine();
            }

            return new Matrix<string, string, TElement>(rowLabels[0], columnLabels[0], data);
        }

        #endregion
    }
}
