#region Copyright information
// 
// Copyright  2005-2013 Yongkee Cho. All rights reserved.
// 
// This code is a part of the Biological Object Library and governed under the terms of the
// GNU Lesser General  Public License (LGPL) version 2.1 which accompanies this distribution.
// For more information on the LGPL, please visit http://bol.codeplex.com/license.
// 
// - Filename: GammaDistribution.cs
// - Author: Yongkee Cho
// - Email: yongkeecho@gmail.com
// - Date Created: 2012-09-06 11:39 AM
// - Last Modified: 2013-01-25 3:59 PM
// 
#endregion
using System;
using System.Collections.Generic;
using System.Linq;
using BOL.Maths.Functions;

namespace BOL.Maths.Distributions
{
    /// <summary>
    /// Represents a gamma distribution with shape parameter, alpha and scale parameter, beta.
    /// <see>Mathematica: ref/GammaDistribution</see>
    /// <see>http://en.wikipedia.org/wiki/Gamma_distribution</see>
    /// </summary>
    public sealed class GammaDistribution : IUnivariateDistribution<double>, IEquatable<GammaDistribution>
    {
        #region Private variables

        private double _alpha;
        private double _beta;

        #endregion

        #region Public properties

        /// <summary>Gets or sets the shape parameter, alpha of a gamma distribution.</summary>
        public double Alpha
        {
            get { return _alpha; }
            set
            {
                if (value <= 0)
                    throw new ArgumentOutOfRangeException("value", "alpha is expected to be a positive real number.");
                
                _alpha = value;
            }
        }

        /// <summary>Gets or sets the scale parameter, beta of a gamma distribution.</summary>
        public double Beta
        {
            get { return _beta; }
            set
            {
                if (value <= 0)
                    throw new ArgumentOutOfRangeException("value", "beta is expected to be a positive real number.");
                
                _beta = value;
            }
        }

        /// <summary>Gets the range of a gamma distribution.</summary>
        public IRange<double> Domain { get { return new Range<double>(0, double.PositiveInfinity); } }

        /// <summary>Gets the mean of a gamma distribution.</summary>
        public double Mean { get { return _alpha * _beta; } }

        /// <summary>Gets the variance of a gamma distribution.</summary>
        public double Variance { get { return _alpha * _beta * _beta; } }

        /// <summary>Gets the median of a gamma distribution.</summary>
        public double Median { get { throw new Exception("Median is undefined for gamma distribution."); } }

        /// <summary>Gets the mode of a gamma distribution.</summary>
        public double Mode
        {
            get
            {
                if (_alpha < 1)
                    throw new Exception("Alpha is less than one.");

                return (_alpha - 1) * _beta;
            }
        }

        /// <summary>Gets the skewness of a gamma distribution.</summary>
        public double Skewness { get { return 2 / Math.Sqrt(_alpha); } }

        /// <summary>Gets the kurtosis of a gamma distribution.</summary>
        public double Kurtosis { get { return 6 / _alpha; } }

        /// <summary>Gets the entropy of a gamma distribution.</summary>
        public double Entropy { get { return _alpha + Math.Log(_beta) + DistributionFunctions.LogGamma(_alpha) + (1 - _alpha) * DistributionFunctions.DiGamma(_alpha); } }

        #endregion

        #region Constructor

        /// <summary>
        /// Instantiates a Gamma distribution.
        /// </summary>
        /// <param name="alpha">shape parameter</param>
        /// <param name="beta">scale parameter</param>
        public GammaDistribution(double alpha, double beta)
        {
            if (alpha <= 0)
                throw new ArgumentOutOfRangeException("alpha", "alpha is expected to be a positive real number.");
            if (beta <= 0)
                throw new ArgumentOutOfRangeException("alpha", "beta is expected to be a positive real number.");
            
            _alpha = alpha;
            _beta = beta;
        }

        #endregion

        #region ICloneable implementation

        public GammaDistribution Clone()
        {
            return new GammaDistribution(_alpha, _beta);
        }

        object ICloneable.Clone()
        {
            return Clone();
        }
        
        #endregion

        #region Public methods

        /// <summary>
        /// Returns the gamma density given value.
        /// </summary>
        /// <param name="value"></param>
        /// <returns>gamma density</returns>
        public double Pdf(double value)
        {
            if (value < 0)
                throw new ArgumentOutOfRangeException("value", "value is expected to be a positive real number.");

            return Math.Exp(-value / _beta) * Math.Pow(value, _alpha - 1) * Math.Pow(_beta, -_alpha) / DistributionFunctions.Gamma(_alpha);   
        }

        /// <summary>
        /// Returns the cummulative gamma density given value.
        /// </summary>
        /// <param name="value"></param>
        /// <returns>cummulative gamma density</returns>
        public double Cdf(double value)
        {
            return DistributionFunctions.IncompleteGamma(_alpha, value);
        }

        /// <summary>
        /// Returns the value of the Gamma distribution for given probability.
        /// </summary>
        /// <param name="p">probability</param>
        /// <returns></returns>
        public double Quantile(double p)
        {
            if (p.Equals(0))
                return 0.0;

            if (p.Equals(1))
                return Double.MaxValue;

            //return FindRoot(probability, shape, 0.0, Double.MAX_VALUE);
            throw new NotImplementedException();
        }

        /// <summary>
        /// Returns a random value from the gamma distribution.
        /// <see>http://people.csail.mit.edu/milch/blog/index.html</see>
        /// <see>Non-Uniform Random Variate Generation, Devroye)  Uses Cheng's rejection algorithm (GB) for k>=1, rejection from Weibull distribution for k in [0, 1].</see>
        /// </summary>
        /// <returns>random value</returns>
        public double Sample(Random rand)
        {
            var accept = false;

            if (_alpha >= 1) // Cheng's algorithm
            {
                var b = (_alpha - Math.Log(4));
                var c = (_alpha + Math.Sqrt(2 * _alpha - 1));
                var lam = Math.Sqrt(2 * _alpha - 1);
                var cheng = (1 + Math.Log(4.5));
                double x;

                do
                {
                    var u = rand.NextDouble();
                    var v = rand.NextDouble();
                    var y = ((1 / lam) * Math.Log(v / (1 - v)));
                    x = (_alpha * Math.Exp(y));
                    var z = (u * v * v);
                    var r = (b + (c * y) - x);
                    if ((r >= ((4.5 * z) - cheng)) || (r >= Math.Log(z)))
                        accept = true;
                } while (!accept);

                return x / _beta;
            }
            else // Weibull algorithm
            {
                var c = (1 / _alpha);
                var d = ((1 - _alpha) * Math.Pow(_alpha, (_alpha / (1 - _alpha))));
                double x;
                do
                {
                    var u = rand.NextDouble();
                    var v = rand.NextDouble();
                    var z = -Math.Log(u); //generating random exponential variates
                    var e = -Math.Log(v);
                    x = Math.Pow(z, c);
                    if ((z + e) >= (d + x))
                        accept = true;
                } while (!accept);

                return x / _beta;
            }
        }

        /// <summary>
        /// Estimates parameters of the gamma distribution using maximum likelihood.
        /// <see>http://en.wikipedia.org/wiki/Gamma_Distribution#Maximum_likelihood_estimation</see>
        /// </summary>
        public void MaximumLikelihoodEstimate(IEnumerable<double> source)
        {
            if (source == null)
                throw new ArgumentNullException("source");

            var list = source.ToList();
            var n = list.Count;
            var mean = list.Average();
            var logMean = list.Where(l => l > 0).Sum(x => Math.Log(x)) / n;
            var s = mean.Equals(0) ? 0 : Math.Log(mean) - logMean;
            Alpha = (3 - s + Math.Sqrt(Math.Pow(s - 3, 2) + 24 * s)) / (12 * s);
            Beta = mean / Alpha;
        }

        /// <summary>
        /// Estimates parameters of the gamma distribution using Bayesian methods.
        /// </summary>
        public void BayesianEstimate(IEnumerable<double> source, params dynamic[] priorParameters)
        {
            throw new NotImplementedException();
        }

        #endregion

        #region IEquatable<GammaDistribution> implementation

        public bool Equals(GammaDistribution other)
        {
            return _alpha.Equals(other._alpha) & _beta.Equals(other._beta);
        }

        #endregion

        #region Object overriden

        public override int GetHashCode()
        {
            return Alpha.GetHashCode() ^ Beta.GetHashCode();
        }

        public override bool Equals(object other)
        {
            if (other == null)
                throw new ArgumentNullException("other");

            if (!(other is GammaDistribution))
                throw new InvalidCastException("The 'other' argument is not a GammaDistribution object.");
            
            return Equals(other as GammaDistribution);
        }

        public override string ToString()
        {
            return String.Format("Gamma(alpha = {0:G4}, beta = {1:G4})", _alpha, _beta);
        }

        #endregion
    }
}
