#region Copyright information
// 
// Copyright  2005-2013 Yongkee Cho. All rights reserved.
// 
// This code is a part of the Biological Object Library and governed under the terms of the
// GNU Lesser General  Public License (LGPL) version 2.1 which accompanies this distribution.
// For more information on the LGPL, please visit http://bol.codeplex.com/license.
// 
// - Filename: InverseGammaDistribution.cs
// - Author: Yongkee Cho
// - Email: yongkeecho@gmail.com
// - Date Created: 2012-09-06 11:39 AM
// - Last Modified: 2013-01-25 3:59 PM
// 
#endregion
using System;
using System.Collections.Generic;
using BOL.Maths.Functions;

namespace BOL.Maths.Distributions
{
    /// <summary>
    /// Represents a inverse gamma distribution with shape parameter, alpha and scale parameter, beta.
    /// <see>Mathematica: ref/InverseGammaDistribution</see>
    /// <see>http://en.wikipedia.org/wiki/Inverse_gamma_distribution</see>
    /// </summary>
    public sealed class InverseGammaDistribution : IUnivariateDistribution<double>, IEquatable<InverseGammaDistribution>
    {
        #region Private variables

        private double _alpha;
        private double _gammaAlpha;
        private double _beta;

        #endregion

        #region Public properties

        /// <summary>Gets or sets the shape parameter, alpha of an inverse gamma distribution.</summary>
        public double Alpha
        {
            get { return _alpha; }
            set
            {
                if (value <= 0)
                    throw new ArgumentOutOfRangeException("value", "alpha is expected to be a positive real number.");
                
                _alpha = value;
                _gammaAlpha = DistributionFunctions.Gamma(_alpha);
            }
        }

        /// <summary>Gets or sets the scale parameter, beta of an inverse gamma distribution.</summary>
        public double Beta
        {
            get { return _beta; }
            set
            {
                if (value <= 0)
                    throw new ArgumentOutOfRangeException("value", "beta is expected to be a positive real number.");
                
                _beta = value;
            }
        }

        /// <summary>Gets the range of an inverse gamma distribution.</summary>
        public IRange<double> Domain { get { return new Range<double>(0.0, double.PositiveInfinity); } }

        /// <summary>Gets the mean of an inverse gamma distribution.</summary>
        public double Mean
        {
            get
            {
                if (_alpha.Equals(1))
                    throw new Exception("Mean is defined for alpha > 1.");
                
                return _beta / (_alpha - 1);
            }
        }

        /// <summary>Gets the median of an inverse gamma distribution.</summary>
        public double Median { get { throw new Exception("Median is undefined for inverse gamma distribution."); } }

        /// <summary>Gets the mode of an inverse gamma distribution.</summary>
        public double Mode { get { return _beta / (_alpha + 1); } }

        /// <summary>Gets the variance of an inverse gamma distribution.</summary>
        public double Variance
        {
            get
            {
                if (_alpha <= 2)
                    throw new Exception("Variance is defined for alpha > 2.");
                
                return _beta * _beta / Math.Pow(_alpha - 1, 2) / (_alpha - 2);
            }
        }

        /// <summary>Gets the skewness of an inverse gamma distribution.</summary>
        public double Skewness
        {
            get
            {
                if (_alpha <= 3)
                    throw new Exception("Skewness is defined for alpha > 3.");

                return 4 * Math.Sqrt(_alpha - 2) / (_alpha - 3);
            }
        }

        /// <summary>Gets the kurtosis of an inverse gamma distribution.</summary>
        public double Kurtosis
        {
            get
            {
                if (_alpha <= 4)
                    throw new Exception("Kurtosis is defined for alpha > 4.");

                return (30 * _alpha - 66) / (_alpha - 3) / (_alpha - 4);
            }
        }

        /// <summary>Gets the entropy of an inverse gamma distribution.</summary>
        public double Entropy { get { return _alpha + Math.Log(_beta * _gammaAlpha) - (1 + _alpha) * DistributionFunctions.DiGamma(_alpha); } }

        #endregion

        #region Constructor

        /// <summary>
        /// Instantiates an inverse gamma distribution.
        /// </summary>
        /// <param name="alpha">shape parameter</param>
        /// <param name="beta">scale parameter</param>
        public InverseGammaDistribution(double alpha, double beta)
        {
            if (alpha <= 0)
                throw new ArgumentOutOfRangeException("alpha", "alpha is expected to be a positive real number.");
            if (beta <= 0)
                throw new ArgumentOutOfRangeException("beta", "beta is expected to be a positive real number.");

            _alpha = alpha;
            _gammaAlpha = DistributionFunctions.Gamma(_alpha);
            _beta = beta;
        }

        #endregion

        #region ICloneable implementation

        public InverseGammaDistribution Clone()
        {
            return new InverseGammaDistribution(_alpha, _beta);
        }

        object ICloneable.Clone()
        {
            return Clone();
        }
        
        #endregion

        #region Public methods

        /// <summary>
        /// Returns the inverse gamma density given value.
        /// </summary>
        /// <param name="value"></param>
        /// <returns>inverse gamma density</returns>
        public double Pdf(double value)
        {
            if (value <= 0)
                throw new ArgumentOutOfRangeException("value", "value is expected to be a positive real number.");

            return Math.Pow(_beta, _alpha) * Math.Pow(value, -_alpha - 1) * Math.Exp(-_beta / value) / _gammaAlpha;
        }

        /// <summary>
        /// Returns the cummulative inverse gamma density given value.
        /// </summary>
        /// <param name="value"></param>
        /// <returns>cummulative inverse gamma density</returns>
        public double Cdf(double value)
        {
            if (value <= 0)
                throw new ArgumentOutOfRangeException("value", "value is expected to be a positive real number.");

            return DistributionFunctions.IncompleteGamma(_alpha, _beta / value) / _gammaAlpha;
        }

        /// <summary>
        /// Returns the value of the inverse gamma distribution for given probability.
        /// </summary>
        /// <param name="p">probability</param>
        /// <returns></returns>
        public double Quantile(double p)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Returns a random value from the inverse gamma distribution.
        /// </summary>
        /// <returns>random value</returns>
        public double Sample(Random r)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Estimates parameters of the inverse gamma distribution using maximum likelihood.
        /// </summary>
        public void MaximumLikelihoodEstimate(IEnumerable<double> source)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Estimates parameters of the inverse gamma distribution using Bayesian methods.
        /// </summary>
        public void BayesianEstimate(IEnumerable<double> source, params dynamic[] priorParameters)
        {
            throw new NotImplementedException();
        }

        #endregion

        #region IEquatable<InverseGammaDistribution> implementation

        public bool Equals(InverseGammaDistribution other)
        {
            return _alpha.Equals(other._alpha) & _beta.Equals(other._beta);
        }

        #endregion

        #region Object overriden

        public override int GetHashCode()
        {
            return Alpha.GetHashCode() ^ Beta.GetHashCode();
        }

        public override bool Equals(object other)
        {
            if (other == null)
                throw new ArgumentNullException("other");

            if (!(other is InverseGammaDistribution))
                throw new InvalidCastException("The 'other' argument is not a InverseGammaDistribution object.");
            
            return Equals(other as InverseGammaDistribution);
        }

        public override string ToString()
        {
            return String.Format("InverseGamma(alpha = {0:G4}, beta = {1:G4})", _alpha, _beta);
        }

        #endregion
    }
}
