#region Copyright information
// 
// Copyright  2005-2013 Yongkee Cho. All rights reserved.
// 
// This code is a part of the Biological Object Library and governed under the terms of the
// GNU Lesser General  Public License (LGPL) version 2.1 which accompanies this distribution.
// For more information on the LGPL, please visit http://bol.codeplex.com/license.
// 
// - Filename: LogLogisticDistribution.cs
// - Author: Yongkee Cho
// - Email: yongkeecho@gmail.com
// - Date Created: 2012-09-06 11:39 AM
// - Last Modified: 2013-01-25 3:59 PM
// 
#endregion
using System;
using System.Collections.Generic;

namespace BOL.Maths.Distributions
{
    /// <summary>
    /// Represents a log-logistic distribution with scale parameter, alpha and shape parameter, beta.
    /// <see>http://en.wikipedia.org/wiki/Loglogistic_distribution</see>
    /// </summary>
    public sealed class LogLogisticDistribution : IUnivariateDistribution<double>, IEquatable<LogLogisticDistribution>
    {
        #region Private variables

        private double _alpha;
        private double _beta;

        #endregion

        #region Public properties

        /// <summary>Gets or sets the scale parameter, alpha of a log-logistic distribution.</summary>
        public double Alpha { get { return _alpha; } set { _alpha = value; } }

        /// <summary>Gets or sets the shape parameter, beta of a log-logistic distribution.</summary>
        public double Beta { get { return _beta; } set { _beta = value; } }

        /// <summary>Gets the range of a log-logistic distribution.</summary>
        public IRange<double> Domain { get { return new Range<double>(0, double.PositiveInfinity); } }

        /// <summary>Gets the mean of a log-logistic distribution.</summary>
        public double Mean { get { return _alpha * Math.PI / _beta / Math.Sin(Math.PI / _beta); } }

        /// <summary>Gets the median of a log-logistic distribution.</summary>
        public double Median { get { return _alpha; } }

        /// <summary>Gets the mode of a log-logistic distribution.</summary>
        public double Mode { get { return (_beta > 1) ? _alpha * Math.Pow((_beta - 1) / (_beta + 1), 1 / _beta) : 0; } }

        /// <summary>Gets the variance of a log-logistic distribution.</summary>
        public double Variance
        {
            get
            {
                var b = Math.PI / _beta;
                if (_beta > 2)
                    return Math.Pow(_alpha, 2) * (2 * b / Math.Sin(2 * b) - Math.Pow(b, 2) / Math.Pow(Math.Sin(b), 2));
                
                throw new NotSupportedException();
            }
        }

        /// <summary>Gets the skewness of a log-logistic distribution.</summary>
        public double Skewness { get { throw new Exception("Skewness is undefined for log-logistic distribution."); } }

        /// <summary>Gets the kurtosis of a log-logistic distribution.</summary>
        public double Kurtosis { get { throw new Exception("Kurtosis is undefined for log-logistic distribution."); } }

        /// <summary>Gets the entropy of a log-logistic distribution.</summary>
        public double Entropy { get { throw new Exception("Entropy is undefined for log-logistic distribution."); } }

        #endregion

        #region Constructor

        /// <summary>
        /// Instantiates a log-logistic distribution.
        /// </summary>
        /// <param name="alpha">scale parameter</param>
        /// <param name="beta">shape parameter</param>
        public LogLogisticDistribution(double alpha, double beta)
        {
            if (alpha <= 0)
                throw new ArgumentOutOfRangeException("alpha", "alpha is expected to be positive");
            if (beta <= 0)
                throw new ArgumentOutOfRangeException("beta", "beta is expected to be positive");

            _alpha = alpha;
            _beta = beta;
        }

        #endregion

        #region ICloneable implementation

        public LogLogisticDistribution Clone()
        {
            return new LogLogisticDistribution(_alpha, _beta);
        }

        object ICloneable.Clone()
        {
            return Clone();
        }

        #endregion

        #region Public methods

        /// <summary>
        /// Returns the log-logistic density given value.
        /// </summary>
        /// <param name="value"></param>
        /// <returns>log-logistic density</returns>
        public double Pdf(double value)
        {
            if (value <= 0)
                throw new ArgumentOutOfRangeException("value", "value is expected to be positive");

            var temp = value / _alpha;
            return (_beta / _alpha) * Math.Pow(temp, _beta - 1) / Math.Pow(1 + Math.Pow(temp, _beta), 2);
        }

        /// <summary>
        /// Returns the cummulative log-logistic density given value
        /// </summary>
        /// <param name="value"></param>
        /// <returns>cummulative log-logistic density</returns>
        public double Cdf(double value)
        {
            if (value <= 0)
                throw new ArgumentOutOfRangeException("value", "value is expected to be positive");

            var temp = Math.Pow(value, _beta);
            return temp / (Math.Pow(_alpha, _beta) + temp);
        }

        /// <summary>
        /// Returns the value of the log-logistic distribution for given probability.
        /// </summary>
        /// <param name="p">probability</param>
        /// <returns></returns>
        public double Quantile(double p)
        {
            if (p.Equals(0))
                return 0.0;
            
            if (p.Equals(1))
                return 1.0;
            
            return _alpha * Math.Pow(p / (1 - p), 1 / _beta);
        }

        /// <summary>
        /// Returns the survival given value
        /// </summary>
        /// <param name="value">time</param>
        /// <returns>survival</returns>
        public double Survival(double value)
        {
            return 1 - Cdf(value);
        }

        /// <summary>
        /// Returns the hazard given value
        /// </summary>
        /// <param name="value">time</param>
        /// <returns>hazard</returns>
        public double Hazard(double value)
        {
            return Pdf(value) / Survival(value);
        }

        /// <summary>
        /// Returns a random value from the log-logistic distribution.
        /// <see>http://blacklen.wordpress.com/2011/04/30/log-logistic-distribution/</see>
        /// </summary>
        /// <returns>random value</returns>
        public double Sample(Random r)
        {
            return Quantile(r.NextDouble());
        }

        /// <summary>
        /// Estimates parameters of the log-logistic distribution using maximum likelihood.
        /// </summary>
        public void MaximumLikelihoodEstimate(IEnumerable<double> source)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Estimates parameters of the log-logistic distribution using Bayesian methods.
        /// </summary>
        public void BayesianEstimate(IEnumerable<double> source, params dynamic[] priorParameters)
        {
            throw new NotImplementedException();
        }

        #endregion

        #region IEquatable<LogLogisticDistribution> implementation

        public bool Equals(LogLogisticDistribution other)
        {
            return _alpha.Equals(other._alpha) & _beta.Equals(other._beta);
        }

        #endregion

        #region Object overriden

        public override int GetHashCode()
        {
            return Alpha.GetHashCode() ^ Beta.GetHashCode();
        }

        public override bool Equals(object other)
        {
            if (other == null)
                throw new ArgumentNullException("other");

            if (!(other is LogLogisticDistribution))
                throw new InvalidCastException("The 'other' argument is not a LogLogisticDistribution object.");
            
            return Equals(other as LogLogisticDistribution);
        }

        public override string ToString()
        {
            return String.Format("Log-logistic(alpha = {0:G4}, beta = {1:G4})", _alpha, _beta);
        }

        #endregion
    }
}
