#region Copyright information
// 
// Copyright  2005-2013 Yongkee Cho. All rights reserved.
// 
// This code is a part of the Biological Object Library and governed under the terms of the
// GNU Lesser General  Public License (LGPL) version 2.1 which accompanies this distribution.
// For more information on the LGPL, please visit http://bol.codeplex.com/license.
// 
// - Filename: LogisticDistribution.cs
// - Author: Yongkee Cho
// - Email: yongkeecho@gmail.com
// - Date Created: 2012-09-06 11:39 AM
// - Last Modified: 2013-01-25 3:59 PM
// 
#endregion
using System;
using System.Collections.Generic;

namespace BOL.Maths.Distributions
{
    /// <summary>
    /// Represents a logistic distribution with location parameter, alpha and scale parameter, beta.
    /// <see>http://en.wikipedia.org/wiki/Logistic_distribution</see>
    /// </summary>
    public sealed class LogisticDistribution : IUnivariateDistribution<double>, IEquatable<LogisticDistribution>
    {
        #region Private variables

        private double _alpha;
        private double _beta;

        #endregion

        #region Public properties

        /// <summary>Gets or sets the location parameter, alpha of a logistic distribution.</summary>
        public double Alpha { get { return _alpha; } set { _alpha = value; } }

        /// <summary>Gets or sets the scale parameter, beta of a logistic distribution.</summary>
        public double Beta
        {
            get { return _beta; }
            set
            {
                if (value <= 0)
                    throw new ArgumentOutOfRangeException("value", "beta is expected to be positive");

                _beta = value;
            }
        }

        /// <summary>Gets the range of a logistic distribution.</summary>
        public IRange<double> Domain { get { return new Range<double>(double.NegativeInfinity, double.PositiveInfinity); } }

        /// <summary>Gets the mean of a logistic distribution.</summary>
        public double Mean { get { return _alpha; } }

        /// <summary>Gets the median of a logistic distribution.</summary>
        public double Median { get { return _alpha; } }

        /// <summary>Gets the mode of a logistic distribution.</summary>
        public double Mode { get { return _alpha; } }

        /// <summary>Gets the variance of a logistic distribution.</summary>
        public double Variance { get { return Math.Pow(Math.PI * _beta, 2) / 3; } }

        /// <summary>Gets the skewness of a logistic distribution.</summary>
        public double Skewness { get { return 0; } }

        /// <summary>Gets the kurtosis of a logistic distribution.</summary>
        public double Kurtosis { get { return 1.2; } }

        /// <summary>Gets the entropy of a logistic distribution.</summary>
        public double Entropy { get { return Math.Log(_beta) + 2; } }

        #endregion

        #region Constructor

        /// <summary>
        /// Instantiates a logistic distribution.
        /// </summary>
        /// <param name="alpha">location parameter</param>
        /// <param name="beta">scale parameter</param>
        public LogisticDistribution(double alpha, double beta)
        {
            if (beta <= 0)
                throw new ArgumentOutOfRangeException("beta", "beta is expected to be positive");

            _alpha = alpha;
            _beta = beta;
        }

        #endregion

        #region ICloneable implementation

        public LogisticDistribution Clone()
        {
            return new LogisticDistribution(_alpha, _beta);
        }

        object ICloneable.Clone()
        {
            return Clone();
        }

        #endregion

        #region Public methods

        /// <summary>
        /// Returns the logistic density given value.
        /// </summary>
        /// <param name="value"></param>
        /// <returns>logistic density</returns>
        public double Pdf(double value)
        {
            var temp = Math.Exp(-(value - _alpha) / _beta);

            return temp / _beta / Math.Pow(temp, 2);
        }

        /// <summary>
        /// Returns the cummulative logistic density given value
        /// </summary>
        /// <param name="value"></param>
        /// <returns>cummulative logistic density</returns>
        public double Cdf(double value)
        {
            return 1 / (1 + Math.Exp(-(value - _alpha) / _beta));
        }

        /// <summary>
        /// Returns the value of the logistic distribution for given probability.
        /// </summary>
        /// <param name="p">probability</param>
        /// <returns></returns>
        public double Quantile(double p)
        {
            if (p.Equals(0))
                return 0.0;
            
            if (p.Equals(1))
                return 1.0;
            
            return _alpha + _beta * Math.Log(p / (1 - p));
        }

        /// <summary>
        /// Returns a random value from the logistic distribution.
        /// <see>http://www.xycoon.com/logistic_random.htm</see>
        /// </summary>
        /// <returns>random value</returns>
        public double Sample(Random r)
        {
            return Quantile(r.NextDouble());
        }

        /// <summary>
        /// Estimates parameters of the logistic distribution using maximum likelihood.
        /// </summary>
        public void MaximumLikelihoodEstimate(IEnumerable<double> source)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Estimates parameters of the logistic distribution using Bayesian methods.
        /// </summary>
        public void BayesianEstimate(IEnumerable<double> source, params dynamic[] priorParameters)
        {
            throw new NotImplementedException();
        }

        #endregion

        #region IEquatable<LogisticDistribution> implementation

        public bool Equals(LogisticDistribution other)
        {
            return _alpha.Equals(other._alpha) & _beta.Equals(other._beta);
        }

        #endregion

        #region Object overriden

        public override int GetHashCode()
        {
            return Alpha.GetHashCode() ^ Beta.GetHashCode();
        }

        public override bool Equals(object other)
        {
            if (other == null)
                throw new ArgumentNullException("other");

            if (!(other is LogisticDistribution))
                throw new InvalidCastException("The 'other' argument is not a LogisticDistribution object.");
            
            return Equals(other as LogisticDistribution);
        }

        public override string ToString()
        {
            return String.Format("Logistic(alpha = {0:G4}, beta = {1:G4})", _alpha, _beta);
        }

        #endregion
    }
}
