#region Copyright information
// 
// Copyright  2005-2013 Yongkee Cho. All rights reserved.
// 
// This code is a part of the Biological Object Library and governed under the terms of the
// GNU Lesser General  Public License (LGPL) version 2.1 which accompanies this distribution.
// For more information on the LGPL, please visit http://bol.codeplex.com/license.
// 
// - Filename: MaxwellBoltzmannDistribution.cs
// - Author: Yongkee Cho
// - Email: yongkeecho@gmail.com
// - Date Created: 2012-09-06 11:39 AM
// - Last Modified: 2013-01-25 3:59 PM
// 
#endregion
using System;
using System.Collections.Generic;
using BOL.Maths.Functions;

namespace BOL.Maths.Distributions
{
    /// <summary>
    /// Represents a Maxwell-Boltzmann distribution with a parameter.
    /// <see>http://en.wikipedia.org/wiki/Maxwell%E2%80%93Boltzmann_distribution</see>
    /// </summary>
    public sealed class MaxwellBoltzmannDistribution : IUnivariateDistribution<double>, IEquatable<MaxwellBoltzmannDistribution>
    {
        #region Private variables

        private double _a;

        #endregion

        #region Public properties

        /// <summary>Gets or sets the a parameter, a of a Maxwell-Boltzmann distribution.</summary>
        public double A { get { return _a; } set { if (value <= 0) throw new ArgumentOutOfRangeException("value", "a is expected to be positive"); _a = value; } }

        /// <summary>Gets the range of a Maxwell-Boltzmann distribution.</summary>
        public IRange<double> Domain { get { return new Range<double>(0, true, double.PositiveInfinity, false); } }

        /// <summary>Gets the mean of a Maxwell-Boltzmann distribution.</summary>
        public double Mean { get { return 2 * _a * Math.Sqrt(2 / Math.PI); } }

        /// <summary>Gets the median of a Maxwell-Boltzmann distribution.</summary>
        public double Median { get { throw new Exception("Median is undefined for Maxwell-Boltzmann distribution."); } }

        /// <summary>Gets the mode of a Maxwell-Boltzmann distribution.</summary>
        public double Mode { get { return Math.Sqrt(2) * _a; } }

        /// <summary>Gets the variance of a Maxwell-Boltzmann distribution.</summary>
        public double Variance { get { return _a * _a * (3 * Math.PI - 8) / Math.PI; } }

        /// <summary>Gets the skewness of a Maxwell-Boltzmann distribution.</summary>
        public double Skewness { get { return 2 * Math.Sqrt(2) * (16 - 5 * Math.PI) / Math.Pow(3 * Math.PI - 8, 3 / 2); } }

        /// <summary>Gets the kurtosis of a Maxwell-Boltzmann distribution.</summary>
        public double Kurtosis { get { return 4 * (-96 + 40 * Math.PI - 3 * Math.PI * Math.PI) / Math.Pow(3 * Math.PI - 8, 2); } }

        /// <summary>Gets the entropy of a Maxwell-Boltzmann distribution.</summary>
        public double Entropy { get { throw new NotImplementedException(); } }

        #endregion

        #region Constructor

        /// <summary>
        /// Instantiates a Maxwell-Boltzmann distribution.
        /// </summary>
        /// <param name="a">a parameter</param>
        public MaxwellBoltzmannDistribution(double a)
        {
            if (a <= 0)
                throw new ArgumentOutOfRangeException("a", "a is expected to be positive");

            _a = a;
        }

        #endregion

        #region ICloneable implementation

        public MaxwellBoltzmannDistribution Clone()
        {
            return new MaxwellBoltzmannDistribution(_a);
        }

        object ICloneable.Clone()
        {
            return Clone();
        }

        #endregion

        #region Public methods

        /// <summary>
        /// Returns the Maxwell-Boltzmann density given value.
        /// </summary>
        /// <param name="value"></param>
        /// <returns>Maxwell-Boltzmann density</returns>
        public double Pdf(double value)
        {
            var temp = value * value;
            return Math.Sqrt(2 / Math.PI) * temp * Math.Exp(-temp / (2 * Math.Pow(_a, 2))) / Math.Pow(_a, 3);
        }

        /// <summary>
        /// Returns the cummulative Maxwell-Boltzmann density given value
        /// </summary>
        /// <param name="value"></param>
        /// <returns>cummulative Maxwell-Boltzmann density</returns>
        public double Cdf(double value)
        {
            return DistributionFunctions.Erf(value / Math.Sqrt(2) / _a) - Math.Sqrt(2 / Math.PI) * value * Math.Exp(- value * value / (2 * _a * _a)) / _a;
        }

        /// <summary>
        /// Returns the value of the Maxwell-Boltzmann distribution for given probability.
        /// </summary>
        /// <param name="p">probability</param>
        /// <returns></returns>
        public double Quantile(double p)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Returns a random value from the Maxwell-Boltzmann distribution.
        /// </summary>
        /// <returns>random value</returns>
        public double Sample(Random r)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Estimates parameters of the Maxwell-Boltzmann distribution using maximum likelihood.
        /// </summary>
        public void MaximumLikelihoodEstimate(IEnumerable<double> source)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Estimates parameters of the Maxwell-Boltzmann distribution using Bayesian methods.
        /// </summary>
        public void BayesianEstimate(IEnumerable<double> source, params dynamic[] priorParameters)
        {
            throw new NotImplementedException();
        }

        #endregion

        #region IEquatable<Maxwell-BoltzmannDistribution> implementation

        public bool Equals(MaxwellBoltzmannDistribution other)
        {
            return _a.Equals(other._a);
        }

        #endregion

        #region Object overriden

        public override int GetHashCode()
        {
            return A.GetHashCode();
        }

        public override bool Equals(object other)
        {
            if (other == null)
                throw new ArgumentNullException("other");

            if (!(other is MaxwellBoltzmannDistribution))
                throw new InvalidCastException("The 'other' argument is not a Maxwell-BoltzmannDistribution object.");
            
            return Equals(other as MaxwellBoltzmannDistribution);
        }

        public override string ToString()
        {
            return String.Format("Maxwell-Boltzmann(a = {0})", _a);
        }

        #endregion
    }
}
