#region Copyright information
// 
// Copyright  2005-2013 Yongkee Cho. All rights reserved.
// 
// This code is a part of the Biological Object Library and governed under the terms of the
// GNU Lesser General  Public License (LGPL) version 2.1 which accompanies this distribution.
// For more information on the LGPL, please visit http://bol.codeplex.com/license.
// 
// - Filename: ClassificationMeasures.cs
// - Author: Yongkee Cho
// - Email: yongkeecho@gmail.com
// - Date Created: 2012-10-11 5:05 PM
// - Last Modified: 2013-01-25 3:59 PM
// 
#endregion
using System;

namespace BOL.Maths.Statistics
{
    /// <summary>
    /// Represents a class to estimate statistical measures of the performance of binary classification test.
    /// </summary>
    public class ClassificationMeasures : IEquatable<ClassificationMeasures>
    {
        #region Protected variables

        private readonly int _tp;
        private readonly int _tn;
        private readonly int _fp;
        private readonly int _fn;

        #endregion

        #region Public properties

        /// <summary>
        /// True positives (TP)
        /// </summary>
        public int TruePositives { get { return _tp; } }

        /// <summary>
        /// True negatives (TN)
        /// </summary>
        public int TrueNegatives { get { return _tn; } }

        /// <summary>
        /// False positives (FP)
        /// </summary>
        public int FalsePositives { get { return _fp; } }

        /// <summary>
        /// False negatives (FN)
        /// </summary>
        public int FalseNegatives { get { return _fn; } }

        /// <summary>
        /// Actual positives = (TP + TN)
        /// </summary>
        public int ActualPositives { get { return _tp + _fn; } }

        /// <summary>
        /// Actual negatives = (TP + TN)
        /// </summary>
        public int ActualNegatives { get { return _tn + _fp; } }

        /// <summary>
        /// Predicted positives = (TP + TN)
        /// </summary>
        public int PredictedPositives { get { return _tp + _fp; } }

        /// <summary>
        /// Predicted negatives = (TP + TN)
        /// </summary>
        public int PredictedNegatives { get { return _tn + _fn; } }

        /// <summary>
        /// Accuracy = (TP + TN) * (TP + FP + TN + FN)
        /// </summary>
        public double Accuracy { get { return (_tp + _tn) / (double)(_tp + _fp + _tn + _fn); } }

        /// <summary>
        /// Precision = TP * (TP + FP)
        /// </summary>
        public double Precision { get { return _tp / (double)(_tp + _fp); } }

        /// <summary>
        /// Recall = TP * (TP + FN)
        /// </summary>
        public double Recall { get { return _tp / (double)(_tp + _fn); } }

        /// <summary>
        /// Specificity = TN / (TN + FP)
        /// </summary>
        public double Specificity { get { return _tn / (double)(_tn + _fp); } }

        /// <summary>
        /// Sensitivity or true positive rate (TPR) = TP / (TP + FN)
        /// </summary>
        public double Sensitivity { get { return _tp / (double)(_tp + _fn); } }

        /// <summary>
        /// False positive rate (alpha) = 1 - Specificity
        /// </summary>
        public double FalsePositiveRate { get { return 1 - Specificity; } }

        /// <summary>
        /// False negative rate (beta) = 1 - Sensitivity
        /// </summary>
        public double FalseNegativeRate { get { return 1 - Sensitivity; } }

        /// <summary>
        /// Positive likelihoood ratio = Sensitivity * (1 - Specificity)
        /// </summary>
        public double PositiveLikelihooodRatio { get { return Sensitivity / FalsePositiveRate; } }

        /// <summary>
        /// Negative likelihoood ratio = (1 - Sensitivity) * Specificity
        /// </summary>
        public double NegativeLikelihooodRatio { get { return FalseNegativeRate / Specificity; } }

        /// <summary>
        /// Positive predictive value = TP * (TP + FP)
        /// </summary>
        public double PositivePredictiveValue { get { return Precision; } }

        /// <summary>
        /// Negative predictive value = TN * (TN + FN)
        /// </summary>
        public double NegativePredictiveValue { get { return _tn / (double)(_tn + _fn); } }

        /// <summary>
        /// False discovery rate = FP / (TP + FP)
        /// </summary>
        public double FalseDiscoveryRate { get { return _fp / (double)(_tp + _fp); } }

        /// <summary>
        /// Matthews correlation coefficient (MCC) = (TP * TN - FP * FN) / SQRT((TP + FP)(TP + FN)(TN + FP)(TN + FN))
        /// </summary>
        public double MatthewsCorrelationCoefficient { get { return (_tp * _tn - _fp * _fn) / Math.Sqrt((_tp + _fp) * (_tp + _fn) * (_tn + _fp) * (_tn + _fn)); } }

        /// <summary>
        /// F1 score = 2TP / (P + P')
        /// </summary>
        public double F1 { get { return 2 * Precision * Recall / (Precision + Recall); } }

        #endregion

        #region Constructor

        public ClassificationMeasures(int tp, int tn, int fp, int fn)
        {
            _tp = tp;
            _tn = tn;
            _fp = fp;
            _fn = fn;
        }

        #endregion

        #region IEquatable<ClassificationMeasures> implementation

        public bool Equals(ClassificationMeasures other)
        {
            return TruePositives.Equals(other.TruePositives) & TrueNegatives.Equals(other.TrueNegatives) & FalsePositives.Equals(other.FalsePositives) & FalseNegatives.Equals(other.FalseNegatives);
        }

        #endregion

        #region Object overriden

        public override int GetHashCode()
        {
            return TruePositives ^ TrueNegatives ^ FalsePositives ^ FalseNegatives;
        }

        public override bool Equals(object other)
        {
            if (other == null)
                throw new ArgumentNullException("other");

            if (!(other is ClassificationMeasures))
                throw new InvalidCastException("The 'other' argument is not a ClassificationMeasures object.");

            return Equals(other as ClassificationMeasures);
        }

        public override string ToString()
        {
            return String.Format("TP={0}, TN={1}, FP={2}, FN={3}", _tp, _tn, _fp, _fn);
        }

        #endregion
    }
}
