#region Copyright information
// 
// Copyright  2005-2013 Yongkee Cho. All rights reserved.
// 
// This code is a part of the Biological Object Library and governed under the terms of the
// GNU Lesser General  Public License (LGPL) version 2.1 which accompanies this distribution.
// For more information on the LGPL, please visit http://bol.codeplex.com/license.
// 
// - Filename: Nucleotide.cs
// - Author: Yongkee Cho
// - Email: yongkeecho@gmail.com
// - Date Created: 2013-01-25 3:25 PM
// - Last Modified: 2013-01-25 3:59 PM
// 
#endregion
using System;
using System.Globalization;

namespace BOL.Models.Seq
{
    /// <summary>
    /// Provides a nucleotide struct.
    /// </summary>
    public struct Nucleotide : ICloneable, IResidue, IEquatable<Nucleotide>
    {
        #region Public properties

        /// <summary>Gets the alphabet character of a nucleic acid.</summary>
        public char Alphabet { get; private set; }

        /// <summary>Gets the acronym of a nucleic acid.</summary>
        public string Acronym { get; private set; }

        /// <summary>Gets the full name of a nucleic acid.</summary>
        public string FullName { get; private set; }

        /// <summary>Gets the types of a nucleic acid.</summary>
        public NucleotideTypes Type { get; private set; }

        /// <summary>Gets an Adenine.</summary>
        public static Nucleotide A { get { return new Nucleotide('A'); } }

        /// <summary>Gets a cytosine.</summary>
        public static Nucleotide C { get { return new Nucleotide('C'); } }

        /// <summary>Gets a guanine.</summary>
        public static Nucleotide G { get { return new Nucleotide('G'); } }

        /// <summary>Gets a thymine.</summary>
        public static Nucleotide T { get { return new Nucleotide('T'); } }

        /// <summary>Gets an uracil.</summary>
        public static Nucleotide U { get { return new Nucleotide('U'); } }

        /// <summary>Gets an unknown nucleic acid.</summary>
        public static Nucleotide X { get { return new Nucleotide('X'); } }

        #endregion

        #region Constructors

        public Nucleotide(char alphabet) : this()
        {
            Alphabet = Char.ToUpper(alphabet);

            switch (Alphabet)
            {
                case 'A': Acronym = "Ade"; FullName = "Adenine"; Type = NucleotideTypes.Purine; break;
                case 'C': Acronym = "Cyt"; FullName = "Cytosine"; Type = NucleotideTypes.Pyrimidine; break;
                case 'G': Acronym = "Gua"; FullName = "Guanine"; Type = NucleotideTypes.Purine; break;
                case 'T': Acronym = "Thy"; FullName = "Thymine"; Type = NucleotideTypes.Pyrimidine; break;
                case 'U': Acronym = "Ura"; FullName = "Uracil"; Type = NucleotideTypes.Pyrimidine; break;
                case 'R': Acronym = "-"; FullName = "A/G"; Type = NucleotideTypes.Purine; break;
                case 'Y': Acronym = "-"; FullName = "C/T"; Type = NucleotideTypes.Pyrimidine; break;
                case 'M': Acronym = "-"; FullName = "A/C"; Type = NucleotideTypes.Purine | NucleotideTypes.Pyrimidine; break;
                case 'K': Acronym = "-"; FullName = "G/T"; Type = NucleotideTypes.Purine | NucleotideTypes.Pyrimidine; break;
                case 'S': Acronym = "-"; FullName = "C/G"; Type = NucleotideTypes.Purine | NucleotideTypes.Pyrimidine; break;
                case 'W': Acronym = "-"; FullName = "A/T"; Type = NucleotideTypes.Purine | NucleotideTypes.Pyrimidine; break;
                case 'B': Acronym = "-"; FullName = "C/G/T"; Type = NucleotideTypes.Purine | NucleotideTypes.Pyrimidine; break;
                case 'D': Acronym = "-"; FullName = "A/G/T"; Type = NucleotideTypes.Purine | NucleotideTypes.Pyrimidine; break;
                case 'H': Acronym = "-"; FullName = "A/C/T"; Type = NucleotideTypes.Purine | NucleotideTypes.Pyrimidine; break;
                case 'V': Acronym = "-"; FullName = "A/C/G"; Type = NucleotideTypes.Purine | NucleotideTypes.Pyrimidine; break;
                case 'N': Acronym = "-"; FullName = "A/C/G/T"; Type = NucleotideTypes.Purine | NucleotideTypes.Pyrimidine; break;
                default: Acronym = "Unk"; FullName = "Unknown"; Type = NucleotideTypes.Unknown; break;
            }
        }

        #endregion

        #region Public method

        public Nucleotide GetHydrogenBondPair()
        {
            switch (Alphabet)
            {
                case 'A': return Nucleotide.C;
                case 'C': return Nucleotide.A;
                case 'G': return Nucleotide.T;
                case 'T': return Nucleotide.G;
                case 'U': return Nucleotide.G;
                default: return Nucleotide.X;
            }
        }

        #endregion

        #region ICloneable implementation

        /// <summary>
        /// Creates an exact copy of current Nucleotide.
        /// </summary>
        /// <returns></returns>
        public Nucleotide Clone()
        {
            return new Nucleotide(Alphabet);
        }

        /// <summary>
        /// Creates an exact copy of current T as an object. 
        /// </summary>
        /// <returns></returns>
        object ICloneable.Clone()
        {
            return Clone();
        }

        #endregion

        #region IEquatable<Nucleotide> implementation

        public bool Equals(Nucleotide other)
        {
            return Alphabet.Equals(other.Alphabet);
        }

        #endregion

        #region Object overriden

        public override int GetHashCode()
        {
            return Alphabet.GetHashCode();
        }

        public override bool Equals(object other)
        {
            if (other == null)
                throw new ArgumentNullException("other");

            if (!(other is Nucleotide))
                throw new InvalidCastException("The 'other' argument is not a Nucleotide object.");

            return Equals((Nucleotide)other);
        }

        /// <summary>
        /// Returns a string that represents the Nucleotide object. 
        /// </summary>
        /// <returns>A string that represents the Nucleotide object.</returns>
        public override string ToString()
        {
            return Alphabet.ToString(CultureInfo.InvariantCulture);
        }

        #endregion
    }
}
