#region Copyright information
// 
// Copyright  2005-2013 Yongkee Cho. All rights reserved.
// 
// This code is a part of the Biological Object Library and governed under the terms of the
// GNU Lesser General  Public License (LGPL) version 2.1 which accompanies this distribution.
// For more information on the LGPL, please visit http://bol.codeplex.com/license.
// 
// - Filename: Organism.cs
// - Author: Yongkee Cho
// - Email: yongkeecho@gmail.com
// - Date Created: 2013-01-25 3:25 PM
// - Last Modified: 2013-01-25 3:59 PM
// 
#endregion
using System;
using System.Globalization;
using System.Runtime.Serialization;

namespace BOL.Models.Taxonomy
{
    /// <summary>
    /// Represents organism class
    /// </summary>
    [DataContract]
    public struct Organism : ICloneable, IIdentifiable<int>, INamed, IEquatable<Organism>, IQualifier
    {
        #region Public properties

        /// <summary>Gets the NCBI taxonomy id.</summary>
        [DataMember]
        public int Id { get; private set; }
        string IIdentifiable.Id { get { return Id.ToString(CultureInfo.InvariantCulture); } } 
        
        //public string IIdentifiable.Id { get; }

        /// <summary>Gets the full name of an organism.</summary>
        [DataMember]
        public string Name { get; set; }

        /// <summary>Gets the acronym of an organism.</summary>
        [DataMember]
        public string Acronym { get; private set; }

        #endregion

        #region Constructors

        public Organism(int id, string fullName, string acronym) : this()
        {
            if (String.IsNullOrEmpty(fullName))
                throw new ArgumentNullException("fullName", "Full name cannot be null or empty.");
            if (String.IsNullOrEmpty(acronym))
                throw new ArgumentNullException("acronym", "Acronym cannot be null or empty.");

            Id = id;
            Name = fullName;
            Acronym = acronym;
        }

        public Organism(string fullName, string acronym) : this(9606, fullName, acronym) { }

        #endregion

        #region ICloneable implementation

        /// <summary>
        /// Creates an exact copy of current Organism.
        /// </summary>
        /// <returns></returns>
        public Organism Clone()
        {
            return new Organism(Id, Name, Acronym);
        }

        /// <summary>
        /// Creates an exact copy of current Organism as an object. 
        /// </summary>
        /// <returns></returns>
        object ICloneable.Clone()
        {
            return Clone();
        }

        #endregion

        #region IQualifier implementation

        public string ToQualifierString()
        {
            return String.Format("/organism={0}", ToString());
        }

        #endregion

        #region IEquatable<Organism> implementation

        public bool Equals(Organism other)
        {
            return Id.Equals(other.Id);
        }

        #endregion

        #region Object overriden

        public override int GetHashCode()
        {
            return Id;
        }

        public override bool Equals(object other)
        {
            return Id.Equals(((Organism)other));
        }

        public override string ToString()
        {
            return Name;
        }

        #endregion
    }
}
